@@ -54,7 +54,7 @@ __all__ = ['PSKC', '__version__']
 
 
 # the version number of the library
-__version__ = '0.5'
+__version__ = '1.0'
 
 
 class PSKC(object):
@@ -72,10 +72,12 @@ class PSKC(object):
 
     def __init__(self, filename=None):
         from pskc.encryption import Encryption
+        from pskc.signature import Signature
         from pskc.mac import MAC
         self.version = None
         self.id = None
         self.encryption = Encryption(self)
+        self.signature = Signature(self)
         self.mac = MAC(self)
         self.devices = []
         if filename is not None:
@@ -86,6 +88,7 @@ class PSKC(object):
 
     @property
     def keys(self):
+        """Provide a list of keys."""
         return tuple(key for device in self.devices for key in device.keys)
 
     def add_device(self, **kwargs):
