@@ -1,7 +1,7 @@
 # encryption.py - module for handling encrypted values
 # coding: utf-8
 #
-# Copyright (C) 2014-2016 Arthur de Jong
+# Copyright (C) 2014-2017 Arthur de Jong
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -27,6 +27,10 @@ The encryption key can be derived using the KeyDerivation class.
 """
 
 
+import os
+import re
+
+
 def algorithm_key_lengths(algorithm):
     """Return the possible key lengths for the configured algorithm."""
     from pskc.exceptions import DecryptionError
@@ -38,18 +42,47 @@ def algorithm_key_lengths(algorithm):
         return [int(algorithm[-7:-4]) // 8]
     elif algorithm.endswith('#tripledes-cbc') or \
             algorithm.endswith('#kw-tripledes'):
-        from Crypto.Cipher import DES3
-        return list(DES3.key_size)
+        return [16, 24]
     elif algorithm.endswith('#kw-aes128') or \
             algorithm.endswith('#kw-aes192') or \
             algorithm.endswith('#kw-aes256'):
         return [int(algorithm[-3:]) // 8]
+    elif (algorithm.endswith('#camellia128-cbc') or
+            algorithm.endswith('#camellia192-cbc') or
+            algorithm.endswith('#camellia256-cbc')):
+        return [int(algorithm[-7:-4]) // 8]
+    elif (algorithm.endswith('#kw-camellia128') or
+            algorithm.endswith('#kw-camellia192') or
+            algorithm.endswith('#kw-camellia256')):
+        return [int(algorithm[-3:]) // 8]
     else:
         raise DecryptionError('Unsupported algorithm: %r' % algorithm)
 
 
+def _decrypt_cbc(algorithm, key, ciphertext, iv=None):
+    """Decrypt the ciphertext and return the plaintext value."""
+    from cryptography.hazmat.backends import default_backend
+    from cryptography.hazmat.primitives import padding
+    from cryptography.hazmat.primitives.ciphers import Cipher, modes
+    from pskc.exceptions import DecryptionError
+    if not iv:
+        iv = ciphertext[:algorithm.block_size // 8]
+        ciphertext = ciphertext[algorithm.block_size // 8:]
+    cipher = Cipher(
+        algorithm(key), modes.CBC(iv), backend=default_backend())
+    decryptor = cipher.decryptor()
+    unpadder = padding.PKCS7(algorithm.block_size).unpadder()
+    try:
+        return unpadder.update(
+            decryptor.update(ciphertext) +
+            decryptor.finalize()) + unpadder.finalize()
+    except ValueError:
+        raise DecryptionError('Invalid padding')
+
+
 def decrypt(algorithm, key, ciphertext, iv=None):
     """Decrypt the ciphertext and return the plaintext value."""
+    from cryptography.hazmat.primitives.ciphers import algorithms
     from pskc.exceptions import DecryptionError
     if key is None:
         raise DecryptionError('No key available')
@@ -60,34 +93,48 @@ def decrypt(algorithm, key, ciphertext, iv=None):
     if algorithm.endswith('#aes128-cbc') or \
             algorithm.endswith('#aes192-cbc') or \
             algorithm.endswith('#aes256-cbc'):
-        from Crypto.Cipher import AES
-        from pskc.crypto import unpad
-        if not iv:
-            iv = ciphertext[:AES.block_size]
-            ciphertext = ciphertext[AES.block_size:]
-        cipher = AES.new(key, AES.MODE_CBC, iv)
-        return unpad(cipher.decrypt(ciphertext), AES.block_size)
+        return _decrypt_cbc(algorithms.AES, key, ciphertext, iv)
     elif algorithm.endswith('#tripledes-cbc'):
-        from Crypto.Cipher import DES3
-        from pskc.crypto import unpad
-        if not iv:
-            iv = ciphertext[:DES3.block_size]
-            ciphertext = ciphertext[DES3.block_size:]
-        cipher = DES3.new(key, DES3.MODE_CBC, iv)
-        return unpad(cipher.decrypt(ciphertext), DES3.block_size)
+        return _decrypt_cbc(algorithms.TripleDES, key, ciphertext, iv)
     elif algorithm.endswith('#kw-aes128') or \
             algorithm.endswith('#kw-aes192') or \
             algorithm.endswith('#kw-aes256'):
         from pskc.crypto.aeskw import unwrap
         return unwrap(ciphertext, key)
-    elif algorithm.endswith('#kw-tripledes'):  # pragma: no branch
+    elif algorithm.endswith('#kw-tripledes'):
         from pskc.crypto.tripledeskw import unwrap
         return unwrap(ciphertext, key)
+    elif (algorithm.endswith('#camellia128-cbc') or
+            algorithm.endswith('#camellia192-cbc') or
+            algorithm.endswith('#camellia256-cbc')):
+        return _decrypt_cbc(algorithms.Camellia, key, ciphertext, iv)
+    elif (algorithm.endswith('#kw-camellia128') or  # pragma: no branch
+            algorithm.endswith('#kw-camellia192') or
+            algorithm.endswith('#kw-camellia256')):
+        from pskc.crypto.aeskw import unwrap
+        return unwrap(ciphertext, key, algorithm=algorithms.Camellia)
     # no fallthrough because algorithm_key_lengths() fails with unknown algo
 
 
+def _encrypt_cbc(algorithm, key, plaintext, iv=None):
+    """Encrypt the provided value with the key using the algorithm."""
+    from cryptography.hazmat.backends import default_backend
+    from cryptography.hazmat.primitives import padding
+    from cryptography.hazmat.primitives.ciphers import Cipher, modes
+    iv = iv or os.urandom(algorithm.block_size // 8)
+    cipher = Cipher(
+        algorithm(key), modes.CBC(iv), backend=default_backend())
+    encryptor = cipher.encryptor()
+    padder = padding.PKCS7(algorithm.block_size).padder()
+    return (
+        iv + encryptor.update(
+            padder.update(plaintext) + padder.finalize()) +
+        encryptor.finalize())
+
+
 def encrypt(algorithm, key, plaintext, iv=None):
     """Encrypt the provided value with the key using the algorithm."""
+    from cryptography.hazmat.primitives.ciphers import algorithms
     from pskc.exceptions import EncryptionError
     if key is None:
         raise EncryptionError('No key available')
@@ -98,27 +145,26 @@ def encrypt(algorithm, key, plaintext, iv=None):
     if algorithm.endswith('#aes128-cbc') or \
             algorithm.endswith('#aes192-cbc') or \
             algorithm.endswith('#aes256-cbc'):
-        from Crypto import Random
-        from Crypto.Cipher import AES
-        from pskc.crypto import pad
-        iv = iv or Random.get_random_bytes(AES.block_size)
-        cipher = AES.new(key, AES.MODE_CBC, iv)
-        return iv + cipher.encrypt(pad(plaintext, AES.block_size))
+        return _encrypt_cbc(algorithms.AES, key, plaintext, iv)
     elif algorithm.endswith('#tripledes-cbc'):
-        from Crypto import Random
-        from Crypto.Cipher import DES3
-        from pskc.crypto import pad
-        iv = iv or Random.get_random_bytes(DES3.block_size)
-        cipher = DES3.new(key, DES3.MODE_CBC, iv)
-        return iv + cipher.encrypt(pad(plaintext, DES3.block_size))
+        return _encrypt_cbc(algorithms.TripleDES, key, plaintext, iv)
     elif algorithm.endswith('#kw-aes128') or \
             algorithm.endswith('#kw-aes192') or \
             algorithm.endswith('#kw-aes256'):
         from pskc.crypto.aeskw import wrap
         return wrap(plaintext, key)
-    elif algorithm.endswith('#kw-tripledes'):  # pragma: no branch
+    elif algorithm.endswith('#kw-tripledes'):
         from pskc.crypto.tripledeskw import wrap
         return wrap(plaintext, key)
+    elif (algorithm.endswith('#camellia128-cbc') or
+            algorithm.endswith('#camellia192-cbc') or
+            algorithm.endswith('#camellia256-cbc')):
+        return _encrypt_cbc(algorithms.Camellia, key, plaintext, iv)
+    elif (algorithm.endswith('#kw-camellia128') or  # pragma: no branch
+            algorithm.endswith('#kw-camellia192') or
+            algorithm.endswith('#kw-camellia256')):
+        from pskc.crypto.aeskw import wrap
+        return wrap(plaintext, key, algorithm=algorithms.Camellia)
     # no fallthrough because algorithm_key_lengths() fails with unknown algo
 
 
@@ -135,30 +181,60 @@ class KeyDerivation(object):
     """
 
     def __init__(self):
-        self.algorithm = None
+        self._algorithm = None
         # PBKDF2 properties
         self.pbkdf2_salt = None
         self.pbkdf2_iterations = None
         self.pbkdf2_key_length = None
-        self.pbkdf2_prf = None
+        self._pbkdf2_prf = None
+
+    @property
+    def algorithm(self):
+        """Provide the key derivation algorithm used."""
+        if self._algorithm:
+            return self._algorithm
+
+    @algorithm.setter
+    def algorithm(self, value):
+        from pskc.algorithms import normalise_algorithm
+        self._algorithm = normalise_algorithm(value)
+
+    @property
+    def pbkdf2_prf(self):
+        """Provide the PBKDF2 pseudorandom function used."""
+        if self._pbkdf2_prf:
+            return self._pbkdf2_prf
+
+    @pbkdf2_prf.setter
+    def pbkdf2_prf(self, value):
+        from pskc.algorithms import normalise_algorithm
+        self._pbkdf2_prf = normalise_algorithm(value)
 
     def derive_pbkdf2(self, password):
-        from Crypto.Protocol.KDF import PBKDF2
-        from pskc.mac import get_hmac
+        from hashlib import pbkdf2_hmac
         from pskc.exceptions import KeyDerivationError
-        prf = None
+        prf = 'sha1'
         if self.pbkdf2_prf:
-            prf = get_hmac(self.pbkdf2_prf)
-            if prf is None:
+            match = re.search(
+                r'^(.*#)?hmac-(?P<hash>[a-z0-9-]+)$', self.pbkdf2_prf)
+            if match:
+                prf = match.group('hash')
+            else:
                 raise KeyDerivationError(
-                    'Pseudorandom function unsupported: %r' %
-                    self.pbkdf2_prf)
+                    'Unsupported PRF: %r' % self.pbkdf2_prf)
         if not all((password, self.pbkdf2_salt, self.pbkdf2_key_length,
-                   self.pbkdf2_iterations)):
+                    self.pbkdf2_iterations)):
             raise KeyDerivationError('Incomplete PBKDF2 configuration')
-        return PBKDF2(
-            password, self.pbkdf2_salt, dkLen=self.pbkdf2_key_length,
-            count=self.pbkdf2_iterations, prf=prf)
+        # force conversion to bytestring on Python 3
+        if not isinstance(password, type(b'')):
+            password = password.encode()  # pragma: no cover (Py3 specific)
+        try:
+            return pbkdf2_hmac(
+                prf, password, self.pbkdf2_salt, self.pbkdf2_iterations,
+                self.pbkdf2_key_length)
+        except ValueError:
+            raise KeyDerivationError(
+                'Pseudorandom function unsupported: %r' % self.pbkdf2_prf)
 
     def derive(self, password):
         """Derive a key from the password."""
@@ -173,11 +249,9 @@ class KeyDerivation(object):
 
     def setup_pbkdf2(self, password, salt=None, salt_length=16,
                      key_length=None, iterations=None, prf=None):
-        from Crypto import Random
-        from pskc.algorithms import normalise_algorithm
-        self.algorithm = normalise_algorithm('pbkdf2')
+        self.algorithm = 'pbkdf2'
         if salt is None:
-            salt = Random.get_random_bytes(salt_length)
+            salt = os.urandom(salt_length)
         self.pbkdf2_salt = salt
         if iterations:
             self.pbkdf2_iterations = iterations
@@ -186,7 +260,7 @@ class KeyDerivation(object):
         if key_length:  # pragma: no branch (always specified)
             self.pbkdf2_key_length = key_length
         if prf:
-            self.pbkdf2_prf = normalise_algorithm(prf)
+            self.pbkdf2_prf = prf
         return self.derive_pbkdf2(password)
 
 
@@ -240,6 +314,18 @@ class Encryption(object):
         from pskc.algorithms import normalise_algorithm
         self._algorithm = normalise_algorithm(value)
 
+    @property
+    def is_encrypted(self):
+        """Test whether the PSKC file requires a decryption key."""
+        from pskc.exceptions import DecryptionError
+        try:
+            for key in self.pskc.keys:
+                key.secret, key.counter, key.time_offset
+                key.time_interval, key.time_drift
+        except DecryptionError:
+            return True
+        return False
+
     def derive_key(self, password):
         """Derive a key from the password."""
         self.key = self.derivation.derive(password)
@@ -277,9 +363,8 @@ class Encryption(object):
         self._setup_encryption(kwargs)
         self.key = kwargs.pop('key', self.key)
         if not self.key:
-            from Crypto import Random
-            self.key = Random.get_random_bytes(kwargs.pop(
-                'key_length', self.algorithm_key_lengths[-1]))
+            self.key = os.urandom(
+                kwargs.pop('key_length', self.algorithm_key_lengths[-1]))
 
     def setup_pbkdf2(self, password, **kwargs):
         """Configure password-based PSKC encryption when writing the file.
